"""Annotate - Annotate functions given signature."""

# Extensively modified by CoolCat467

# Extensively modified from https://github.com/dropbox/pyannotate/blob/master/pyannotate_tools/annotations/parse.py

# Pyannotate is licensed under the terms of the Apache License, Version 2.0,
# reproduced below.
#   Copyright (c) 2017 Dropbox, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from __future__ import annotations

__title__ = "Annotate"
__author__ = "CoolCat467"
__license__ = "Apache License, Version 2.0"


import ast
import re
import tokenize
from collections.abc import Callable, Collection, Sequence
from typing import Any, Final, NoReturn

BUILTINS: Final = {"List", "Set", "Type", "Dict", "Tuple"}


class ParseError(Exception):
    """Raised on any type comment parse error.

    The 'comment' attribute contains the comment that produced the error.
    """

    def __init__(self, comment: str | None = None) -> None:
        """Initialize with comment."""
        if comment is None:
            comment = ""
        elif hasattr(self, "add_note"):  # New in Python 3.11
            self.add_note(comment)
        super().__init__(comment)


class Token:
    """Base token."""

    __slots__ = ("text",)

    def __init__(self, text: str | None = None) -> None:
        """Initialize self with text."""
        self.text = text

    def __repr__(self) -> str:
        """Get reprentation for this token."""
        value = "" if self.text is None else f"{self.text!r}"
        return f"{self.__class__.__name__}({value})"


class DottedName(Token):
    """An identifier token, such as 'List', 'int' or 'package.name'."""

    __slots__ = ()


class Separator(Token):
    """A separator or punctuator token such as '(', '[' or '->'."""

    __slots__ = ()


class End(Token):
    """A token representing the end of a type comment."""

    __slots__ = ()

    def __init__(self) -> None:
        """Initialize superclass with no arguments."""
        super().__init__()


def tokenize_annotation(txt: str) -> list[Token]:
    """Translate a type comment into a list of tokens."""
    original = txt
    txt = txt.replace("?", "")
    tokens: list[Token] = []
    while True:
        if not txt:
            tokens.append(End())
            return tokens
        if txt[0] in {" ", "\n"}:
            txt = txt[1:]
        elif txt[0] in "()[],*":
            tokens.append(Separator(txt[0]))
            txt = txt[1:]
        elif txt[:2] == "->":
            tokens.append(Separator("->"))
            txt = txt[2:]
        elif txt[:3] == "...":
            tokens.append(DottedName("..."))
            txt = txt[3:]
        else:
            match_ = re.match(r"[-\w`]+(\s*(\.|:)\s*[-/\w]*)*", txt)
            if not match_:
                raise ParseError(f"Could not parse {txt!r} from {original!r}")
            fullname = match_.group(0)
            size = len(fullname)
            fullname = fullname.replace(" ", "")
            if "`" in fullname:
                fullname = fullname.split("`")[0]
            # pytz creates classes with the name of the timezone being used:
            # https://github.com/stub42/pytz/blob/f55399cddbef67c56db1b83e0939ecc1e276cf42/src/pytz/tzfile.py#L120-L123
            # This causes pyannotate to crash as it's invalid to have a class
            # name with a `/` in it (e.g. "pytz.tzfile.America/Los_Angeles")
            if fullname.startswith("pytz.tzfile."):
                fullname = "datetime.tzinfo"
            if "-" in fullname or "/" in fullname:
                print(f"{__file__}: {fullname} -> Any")
                # Not a valid Python name; there are many places that
                # generate these, so we just substitute Any rather
                # than crashing.
                fullname = "Any"
            tokens.append(DottedName(fullname))
            txt = txt[size:]


def get_line_indent(text: str, char: str = " ") -> int:
    """Return line indent."""
    for idx, cur in enumerate(text.split(char)):
        if cur != "":
            return idx
    return 0


def deindent(level: int, text: str) -> str:
    """Undo indent on text by level of characters."""
    prefix = " " * level
    return "\n".join(line.removeprefix(prefix) for line in text.splitlines())


def tokenize_definition(
    start_line: int,
    get_line: Callable[[int], str],
) -> tuple[ast.FunctionDef, int]:
    """Return Function Definition AST and number of lines after start."""
    current_line_no = start_line
    definition_lines = []

    def read_line() -> str:
        """Incremental wrapper for get_line."""
        nonlocal current_line_no
        value = get_line(current_line_no)
        definition_lines.append(value)
        current_line_no += 1
        return value

    hasdef = False  # Do we have a definition token?
    defstart = False  # Has definition started?
    brackets = 0  # Current number of brackets in use
    typedef = 0  # Brackets in type definition (truthy if in type definition)
    default_arg = 0  # Brackets in default argument value

    # For each token tokenizer module finds
    for token in tokenize.generate_tokens(read_line):
        string = token.string
        token_name = tokenize.tok_name[token.type]
        if token_name == "NAME":  # noqa: S105
            if string == "def":  # Error if more than one in case invalid start
                if hasdef:
                    raise ParseError(
                        "Did not expect second definition keyword",
                    )
                hasdef = True  # Have definition now
        elif token_name == "OP":  # noqa: S105
            if string in "([{":
                # We have traveled in an bracket level
                brackets += 1
                # If we have a definition and this is open parenthesis,
                # we are starting definition
                if not defstart and hasdef and string == "(":
                    defstart = True
                # If doing a type definition, we have typedef level
                if typedef:
                    typedef += 1
                if default_arg:
                    default_arg += 1
            elif string in ")]}":
                # Left a bracket level
                brackets -= 1
                if typedef:  # Maybe left a type def level too
                    typedef -= 1
                if default_arg:
                    default_arg -= 1
            elif string == ",":
                # If exactly one level, leaving type definition land
                if typedef == 1:
                    typedef = 0
                if default_arg == 1:
                    default_arg = 0
            elif string == "->":
                # We are defining a return type
                typedef = 1
            elif string == ":":
                if defstart and not brackets:
                    # Ran out of brackets so done defining function
                    typedef = 0
                    break
                typedef = 1
            elif string == "=":  # Defining a default definition
                typedef = 0  # shouldn't be defining type but make sure
                default_arg = 1
    definition = "".join(definition_lines).rstrip() + " ..."
    indent_level = get_line_indent(definition)
    definition = deindent(indent_level, definition)
    print(f"[{__title__}] DEBUG: {definition = }")

    code = ast.parse(definition, type_comments=True)
    ast_def = code.body[0]
    if not isinstance(ast_def, ast.FunctionDef):
        raise ParseError(f"Expected FunctionDef, got {type(ast_def)}")
    # print(f"{ast.unparse(ast_def)}")
    return ast_def, current_line_no - start_line


def get_type_repr(name: str) -> str:
    """Get representation of name."""
    for seperator in (".", ":"):
        if seperator in name:
            module, text = name.split(seperator, 1)
            if module in ("typing", "mypy_extensions"):
                if text in BUILTINS:
                    return text.lower()
                return text
    return name


def get_typevalue_repr(typevalue: TypeValue) -> str:
    """Get representation of ClassType."""
    name = get_type_repr(typevalue.name)
    if name in BUILTINS:
        name = name.lower()
    args = []
    for arg in typevalue.args:
        args.append(get_typevalue_repr(arg))
    if not args:
        return name
    if name == "Union":
        return " | ".join(args)
    values = ", ".join(args)
    return f"{name}[{values}]"


def typevalue_as_ast(typevalue: TypeValue) -> ast.Name | ast.Subscript:
    """Convert TypeValue into AST Name or Subscript."""
    name = get_type_repr(typevalue.name)
    if not typevalue.args:
        return ast.Name(name, ast.Load())
    slice_: ast.Tuple | ast.Name | ast.Subscript
    if len(typevalue.args) > 1:
        type_args = [typevalue_as_ast(v) for v in typevalue.args]
        slice_ = ast.Tuple(type_args, ast.Load())
    else:
        slice_ = typevalue_as_ast(typevalue.args[0])

    return ast.Subscript(ast.Name(name, ast.Load()), slice_, ast.Load())


class TypeValue:
    """A type value, potentially collection of multiple."""

    __slots__ = ("name", "args")

    def __init__(
        self,
        name: str,
        args: Sequence[TypeValue] | None = None,
    ) -> None:
        """Set up name and arguments."""
        self.name = name
        if args:
            self.args = tuple(args)
        else:
            self.args = ()

    def __repr__(self) -> str:
        """Get representation of self."""
        return f"TypeValue({self.name!r}, {self.args!r})"

    def __str__(self) -> str:
        """Get type value representation of self."""
        return get_typevalue_repr(self)


def list_or(values: Collection[str]) -> str:
    """Return comma separated listing of values joined with ` or `."""
    if len(values) <= 2:
        return " or ".join(values)
    copy = list(values)
    copy[-1] = f"or {copy[-1]}"
    return ", ".join(copy)


class Parser:
    """Implementation of the type comment parser."""

    __slots__ = ("tokens", "i")

    def __init__(self, tokens: list[Token]) -> None:
        """Initialize with tokens list."""
        self.tokens = tokens
        self.i = 0

    def parse_type_list(self) -> list[TypeValue]:
        """Parse comma separated type list."""
        types = []
        while self.lookup() not in (")", "]"):
            typ = self.parse_type()
            types.append(typ)
            string = self.lookup()
            if string == ",":
                self.expect(",")
            elif string not in {")", "]"}:
                self.fail(f"Expected ) or ], got {string!r}")
        return types

    def parse_union_list(self) -> list[TypeValue]:
        """Parse | separated union list."""
        types = []
        while True:
            typ = self.parse_type()
            types.append(typ)
            if self.lookup() != "|":
                return types
            self.expect("|")

    def parse_single(self) -> TypeValue:
        """Parse single, does not look for | unions. Do not use directly."""
        if self.lookup() == "[":
            # Callable[[...], ...]
            #          ^^^^^
            self.expect("[")
            args = self.parse_type_list()
            self.expect("]")
            return TypeValue("", args)
        token = self.expect_type(DottedName)
        if token.text == "Any":
            return TypeValue("Any")
        if token.text == "mypy_extensions.NoReturn":
            return TypeValue("NoReturn")
        if token.text == "typing.NoReturn":
            return TypeValue("NoReturn")
        if token.text == "Tuple":
            self.expect("[")
            args = self.parse_type_list()
            self.expect("]")
            return TypeValue(token.text, args)
        if token.text == "Union":
            self.expect("[")
            items = self.parse_type_list()
            self.expect("]")
            if len(items) == 1:
                return items[0]
            if len(items) == 0:
                self.fail("No items in Union")
            return TypeValue(token.text, items)
        if token.text == "lambda":
            lambda_data = self.parse_lambda()
            return TypeValue(f"{token.text} {lambda_data}")
        if self.lookup() == "[":
            self.expect("[")
            args = self.parse_type_list()
            self.expect("]")
            if token.text == "Optional" and len(args) == 1:
                return TypeValue("Union", [args[0], TypeValue("None")])
            if token.text is None:
                token.text = "None"
            return TypeValue(token.text, args)
        if token.text is None:
            token.text = "None"
        return TypeValue(token.text)

    def parse_lambda_arguments(self) -> list[str]:
        """Parse lambda arguments."""
        arguments = []
        while self.lookup() != ":":
            typ = self.expect_type(DottedName)
            assert typ.text is not None
            arguments.append(typ.text)
            string = self.lookup()
            if string == ",":
                self.expect(",")
            elif string not in {",", ":"}:
                self.fail(f"Expected , or :, got {string!r}")
        return arguments

    def parse_lambda(self) -> str:
        """Parse lambda expression."""
        arguments = ", ".join(self.parse_lambda_arguments())
        self.expect(":")
        value = f"{arguments}: "

        brackets = 0
        while not isinstance(self.peek(), End):
            if self.lookup() == "," and not brackets:
                break
            token = self.next()

            assert token.text is not None
            value += token.text

            if not isinstance(token, DottedName | Separator):
                value += " "

            if token.text in {"(", "[", "{"}:
                brackets += 1
            elif token.text in {")", "]", "}"}:
                brackets -= 1
                if not brackets:
                    break
        return value

    def parse_type(self) -> TypeValue:
        """Parse type including | unions."""
        type_value = self.parse_single()
        if self.lookup() == "|":
            self.expect("|")
            args = [type_value, *self.parse_union_list()]
            return TypeValue("Union", args)
        return type_value

    def fail(self, error: str | None) -> NoReturn:
        """Raise parse error."""
        raise ParseError(error)

    def peek(self) -> Token:
        """Peek at next token."""
        if self.i >= len(self.tokens):
            self.fail("Ran out of tokens")
        return self.tokens[self.i]

    def next(self) -> Token:  # noqa: A003
        """Get next token."""
        token = self.peek()
        self.i += 1
        return token

    def expect(self, text: str) -> None:
        """Expect next token text to be text."""
        got = self.next().text
        if got != text:
            self.fail(f"Expected {text!r}, got {got!r}")

    def expect_type(
        self,
        token_type: type[Token] | tuple[type[Token], ...],
    ) -> Token:
        """Expect next token to be instance of token_type. Return token."""
        token = self.next()
        if not isinstance(token, token_type):
            if hasattr(token_type, "__iter__"):
                assert isinstance(token_type, tuple)
                expect_str = list_or(
                    [f"{cls.__name__!r}" for cls in token_type],
                )
            else:
                expect_str = f"{token_type.__name__!r}"
            self.fail(f"Expected {expect_str}, got {token}")
        return token

    def lookup(self) -> str:
        """Peek at next token and return it's text."""
        value = self.peek().text
        if value is None:
            return "None"
        return value

    def back(self) -> None:
        """Go back one token."""
        self.i -= 1

    def rest_tokens(self) -> list[Token]:
        """Return all tokens not processed."""
        return self.tokens[self.i :]


def get_annotation(
    annotation: dict[str, Any],
    get_line: Callable[[int], str],
) -> tuple[str, int]:
    """Return annotation and end line."""
    # print(f"[DEBUG] Annotate: {annotation = }\n")

    # Get definition tokens
    try:
        func_ast, line_count = tokenize_definition(
            annotation["line"],
            get_line,
        )
    except ParseError:
        print(f"Could not tokenize definition\n{annotation = }")
        raise
    except EOFError as exc:
        raise ParseError(
            "Reached End of File, expected end of definition",
        ) from exc

    # Get the argument and return tokens from signature
    signature = annotation["signature"]
    arg_tokens = [tokenize_annotation(arg) for arg in signature["arg_types"]]
    return_tokens = tokenize_annotation(signature["return_type"])

    arg_idx = 0
    for attr in ("posonlyargs", "args", "vararg", "kwonlyargs", "kwarg"):
        arg_list = getattr(func_ast.args, attr, [])
        if not arg_list:
            continue
        for argument in arg_list:
            if argument.annotation is not None:
                arg_idx += 1
                continue
            parser = Parser(arg_tokens[arg_idx])
            type_value = parser.parse_type()
            argument.annotation = typevalue_as_ast(type_value)
            arg_idx += 1
    if func_ast.returns is None:
        parser = Parser(return_tokens)
        type_value = parser.parse_type()
        func_ast.returns = typevalue_as_ast(type_value)

    new_lines = "\n".join(ast.unparse(func_ast).splitlines()[:-1])

    return new_lines, line_count


def run(annotation: dict[str, str]) -> None:
    """Run test of module."""

    def get_line(line: int) -> str:
        path = annotation["path"]
        assert isinstance(path, str), "Path must be string"
        with open(path, encoding="utf-8") as files:
            for line_no, line_text in enumerate(files):
                if line_no == line - 1:
                    return line_text
            raise EOFError

    print(f"{get_annotation(annotation, get_line)[0]!r}")


if __name__ == "__main__":
    print(f"{__title__}\nProgrammed by {__author__}.\n")
    # annotation = {
    #
    # }
    # run(annotation)



"""Client for mypy daemon mode - Modified version of mypy.dmypy.client.

This manages a daemon process which keeps useful state in memory
rather than having to read it back from disk on each run.
"""

# Modified by CoolCat467
# Original at https://github.com/python/mypy/blob/master/mypy/dmypy/client.py
# Original retrieved November 24th 2022
# Last updated to match original on March 11th 2023

# Mypy (and mypyc) are licensed under the terms of the MIT license,
# reproduced below.
#
#    The MIT License
#
#    Copyright (c) 2012-2022 Jukka Lehtosalo and contributors
#    Copyright (c) 2015-2022 Dropbox, Inc.
#
#    Permission is hereby granted, free of charge, to any person obtaining a
#    copy of this software and associated documentation files (the "Software"),
#    to deal in the Software without restriction, including without limitation
#    the rights to use, copy, modify, merge, publish, distribute, sublicense,
#    and/or sell copies of the Software, and to permit persons to whom the
#    Software is furnished to do so, subject to the following conditions:
#
#    The above copyright notice and this permission notice shall be included in
#    all copies or substantial portions of the Software.
#
#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#    DEALINGS IN THE SOFTWARE.

from __future__ import annotations

__title__ = "Mypy Daemon Client"
__license__ = "MIT"

import contextlib
import io
import json
import os
import time
from collections.abc import Sequence
from typing import TYPE_CHECKING, TypedDict, cast

import trio
from mypy.dmypy_os import alive as _alive, kill as _kill
from mypy.dmypy_server import (
    Server as _Server,
    daemonize as _daemonize,
    process_start_options as _process_start_options,
)
from mypy.ipc import IPCClient as _IPCClient, IPCException as _IPCException
from mypy.version import __version__

if TYPE_CHECKING:
    from typing_extensions import NotRequired


def debug(message: str) -> None:
    """Print debug message."""
    # TODO: Censor username/user files
    print(f"\n[{__title__}] DEBUG: {message}")


class Response(TypedDict):
    """Response dictionary from dmypy."""

    platform: NotRequired[str]
    python_version: NotRequired[str]
    out: NotRequired[str]
    err: NotRequired[str]
    stdout: NotRequired[str]
    stderr: NotRequired[str]
    error: NotRequired[str]
    memory_psutil_missing: NotRequired[str]
    memory_rss_mib: NotRequired[float]
    memory_vms_mib: NotRequired[float]
    memory_maxrss_mib: NotRequired[float]
    restart: NotRequired[str]
    status: NotRequired[int]
    stats: NotRequired[object]


async def _receive(connection: _IPCClient) -> Response:
    """Receive JSON data from a connection until EOF.

    Raise OSError if the data received is not valid JSON or if it is
    not a dict.
    """
    async_connection = trio.wrap_file(cast(io.BytesIO, connection))
    bdata: bytes = await async_connection.read()
    if not bdata:
        raise OSError("No data received")
    try:
        data = json.loads(bdata.decode("utf8"))
    except Exception as ex:
        raise OSError("Data received is not valid JSON") from ex
    if not isinstance(data, dict):
        raise OSError(f"Data received is not a dict (got {type(data)})")
    return cast(Response, data)


class BadStatusError(Exception):

    """Exception raised when there is something wrong with the status file.

    For example:
    - No status file found
    - Status file malformed
    - Process whose process id is in the status file does not exist
    """


# Client-side infrastructure.


def _read_status(status_file: str) -> dict[str, object]:
    """Read status file.

    Raise BadStatusError if the status file doesn't exist or contains
    invalid JSON or the JSON is not a dict.
    """
    if not os.path.isfile(status_file):
        raise BadStatusError("No status file found")
    with open(status_file, encoding="utf-8") as file:
        try:
            data = json.load(file)
        except Exception as ex:
            raise BadStatusError("Malformed status file (not JSON)") from ex
    if not isinstance(data, dict):
        raise BadStatusError("Invalid status file (not a dict)")
    return data


def _check_status(data: dict[str, object]) -> tuple[int, str]:
    """Check if the process is alive.

    Return (process id, connection_name) on success.

    Raise BadStatusError if something's wrong.
    """
    if "pid" not in data:
        raise BadStatusError("Invalid status file (no pid field)")
    pid = data["pid"]
    if not isinstance(pid, int):
        raise BadStatusError("pid field is not an int")
    if not _alive(pid):
        raise BadStatusError("Daemon has died")
    if "connection_name" not in data:
        raise BadStatusError("Invalid status file (no connection_name field)")
    connection_name = data["connection_name"]
    if not isinstance(connection_name, str):
        raise BadStatusError("connection_name field is not a string")
    return pid, connection_name


def get_status(status_file: str) -> tuple[int, str]:
    """Read status file and check if the process is alive.

    Return (process id, connection_name) on success.

    Raise BadStatusError if something's wrong.
    """
    data = _read_status(status_file)
    return _check_status(data)


async def request(
    status_file: str,
    command: str,
    *,
    timeout: int | None = None,
    **kwds: object,
) -> Response:
    """Send a request to the daemon.

    Return the JSON dict with the response.

    Raise BadStatusError if there is something wrong with the status file
    or if the process whose process id is in the status file has died.

    Return {'error': <message>} if an IPC operation or receive()
    raised OSError.  This covers cases such as connection refused or
    closed prematurely as well as invalid JSON received.
    """
    args = dict(kwds)
    args["command"] = command
    # Tell the server whether this request was initiated from a
    # human-facing terminal, so that it can format the type checking
    # output accordingly.
    args["is_tty"] = False
    args["terminal_width"] = 80
    bdata = json.dumps(args).encode("utf8")
    _, name = get_status(status_file)
    try:
        with _IPCClient(name, timeout) as client:
            async_client = trio.wrap_file(cast(io.BytesIO, client))
            await async_client.write(bdata)
            return await _receive(client)
    except (OSError, _IPCException, ValueError) as err:
        return {"error": str(err)}


def is_running(status_file: str) -> bool:
    """Check if the server is running cleanly."""
    try:
        get_status(status_file)
    except BadStatusError:
        return False
    return True


async def stop(status_file: str) -> Response:
    """Stop daemon via a 'stop' request."""
    return await request(status_file, "stop", timeout=5)


async def _wait_for_server(status_file: str, timeout: float = 5.0) -> bool:
    """Wait until the server is up. Return False if timed out."""
    endtime = time.time() + timeout
    while time.time() < endtime:
        try:
            data = _read_status(status_file)
        except BadStatusError:
            # If the file isn't there yet, retry later.
            await trio.sleep(0.1)
            continue
        # If the file's content is bogus or the process is dead, fail.
        try:
            _check_status(data)
        except BadStatusError:
            return False
        return True
    return False


async def _start_server(
    status_file: str,
    *,
    flags: list[str],
    daemon_timeout: int | None = None,
    allow_sources: bool = False,
    log_file: str | None = None,
) -> bool:
    """Start the server and wait for it. Return False if error starting."""
    start_options = _process_start_options(flags, allow_sources)
    if (
        _daemonize(
            start_options,
            status_file,
            timeout=daemon_timeout,
            log_file=log_file,
        )
        != 0
    ):
        return False
    return await _wait_for_server(status_file)


async def restart(
    status_file: str,
    *,
    flags: list[str],
    daemon_timeout: int | None = 0,
    allow_sources: bool = False,
    log_file: str | None = None,
) -> bool:
    """Restart daemon (it may or may not be running; but not hanging).

    Returns False if error starting.
    """
    # Bad or missing status file or dead process; good to start.
    with contextlib.suppress(BadStatusError):
        await stop(status_file)
    return await _start_server(
        status_file,
        flags=flags,
        daemon_timeout=daemon_timeout,
        allow_sources=allow_sources,
        log_file=log_file,
    )


# Action functions


async def start(
    status_file: str,
    *,
    flags: list[str],
    daemon_timeout: int = 0,
    allow_sources: bool = False,
    log_file: str | None = None,
) -> bool:
    """Start daemon if not already running.

    Returns False if error starting / already running.
    """
    if not is_running(status_file):
        # Bad or missing status file or dead process; good to start.
        return await _start_server(
            status_file,
            flags=flags,
            daemon_timeout=daemon_timeout,
            allow_sources=allow_sources,
            log_file=log_file,
        )
    return False


async def status(
    status_file: str,
    *,
    timeout: int = 5,
    fswatcher_dump_file: str | None = None,
) -> Response:
    """Ask daemon to return status."""
    return await request(
        status_file,
        "status",
        timeout=timeout,
        fswatcher_dump_file=fswatcher_dump_file,
    )


async def run(
    status_file: str,
    *,
    flags: list[str],
    timeout: int | None = None,
    daemon_timeout: int = 0,
    log_file: str | None = None,
    export_types: bool = False,
) -> Response:
    """Do a check, starting (or restarting) the daemon as necessary.

    Restarts the daemon if the running daemon reports that it is
    required (due to a configuration change, for example).
    """
    # Start if missing status file or dead process
    await start(
        status_file=status_file,
        flags=flags,
        daemon_timeout=daemon_timeout,
        allow_sources=True,
        log_file=log_file,
    )
    response = await request(
        status_file,
        "run",
        timeout=timeout,
        version=__version__,
        args=flags,
        export_types=export_types,
    )
    # If the daemon signals that a restart is necessary, do it
    if "restart" in response:
        debug(f'Restarting: "{response["restart"]}"')
        await restart(
            status_file,
            flags=flags,
            daemon_timeout=timeout,
            allow_sources=True,
            log_file=log_file,
        )
        return await request(
            status_file,
            "run",
            timeout=timeout,
            version=__version__,
            args=flags,
            export_types=export_types,
        )
    return response


def kill(status_file: str) -> bool:
    """Kill daemon process with SIGKILL. Return True if killed."""
    pid = get_status(status_file)[0]
    try:
        _kill(pid)
    except OSError as ex:
        print(ex)
        return False
    return True


async def check(
    status_file: str,
    files: Sequence[str],
    *,
    timeout: int | None = None,
    export_types: bool = False,
) -> Response:
    """Ask the daemon to check a list of files."""
    return await request(
        status_file,
        "check",
        timeout=timeout,
        files=files,
        export_types=export_types,
    )


async def recheck(
    status_file: str,
    export_types: bool,
    *,
    timeout: int | None = None,
    remove: list[str] | None = None,
    update: list[str] | None = None,
) -> Response:
    """Ask the daemon to recheck the previous list of files.

    If at least one of --remove or --update is given, the server will
    update the list of files to check accordingly and assume that any other
    files are unchanged.  If none of these flags are given, the server will
    call stat() on each file last checked to determine its status.

    Files given in --update ought to exist.  Files given in --remove need not
    exist; if they don't they will be ignored.
    The lists may be empty but oughtn't contain duplicates or overlap.

    NOTE: The list of files is lost when the daemon is restarted.
    """
    if remove is not None or update is not None:
        return await request(
            status_file,
            "recheck",
            timeout=timeout,
            export_types=export_types,
            remove=remove,
            update=update,
        )
    return await request(
        status_file,
        "recheck",
        timeout=timeout,
        export_types=export_types,
    )


async def inspect(
    status_file: str,
    location: str,  # line:col
    show: str = "type",  # type, attrs, definition
    *,
    timeout: int | None = None,
    verbosity: int = 0,
    limit: int = 0,
    include_span: bool = False,
    include_kind: bool = False,
    include_object_attrs: bool = False,
    union_attrs: bool = False,
    force_reload: bool = False,
) -> Response:
    """Ask daemon to print the type of an expression."""
    return await request(
        status_file,
        "inspect",
        timeout=timeout,
        show=show,
        location=location,
        verbosity=verbosity,
        limit=limit,
        include_span=include_span,
        include_kind=include_kind,
        include_object_attrs=include_object_attrs,
        union_attrs=union_attrs,
        force_reload=force_reload,
    )


async def suggest(
    status_file: str,
    function: str,
    do_json: bool,
    *,
    timeout: int | None = None,
    callsites: bool = False,
    no_errors: bool = False,
    no_any: bool = False,
    flex_any: float | None = None,
    use_fixme: str | None = None,
    max_guesses: int | None = 64,
) -> Response:
    """Ask the daemon for a suggested signature."""
    return await request(
        status_file,
        "suggest",
        timeout=timeout,
        function=function,
        json=do_json,
        callsites=callsites,
        no_errors=no_errors,
        no_any=no_any,
        flex_any=flex_any,
        use_fixme=use_fixme,
        max_guesses=max_guesses,
    )


async def hang(status_file: str, *, timeout: int = 1) -> Response:
    """Hang for 100 seconds, as a debug hack."""
    if not isinstance(timeout, int):
        raise ValueError("Timeout must be an integer!")
    return await request(status_file, "hang", timeout=timeout)


def do_daemon(
    status_file: str,
    flags: list[str],
    daemon_timeout: int | None = None,
) -> None:
    """Serve requests in the foreground."""
    options = _process_start_options(flags, allow_sources=False)
    _Server(options, status_file, timeout=daemon_timeout).serve()
